      SUBROUTINE arb_wvn(lambda_h, lambda_t, nlambda_h, nlambda_t,
     $      nlev, rad, rad_h, maxlev )
*
*BOD
*DESCRIPTION
* Interpolates the radiation field rad at resolution lambda_t to 
* the resolution lambda_h using qubic splines. 
*USAGE
**FORTRAN
*      CALL arb_wvn(lambda_h, lambda_t, nlambda_h, nlambda_t,
*     $      nlev, rad, rad_h)
**PERL
*INPUT
**lambda_h
*     real array(*)
*     High resolution wavelengths that rad will be interpolated to.
**lambda_t
*     real array(*)
*     Low resolution wavelengths at which rad is given.
**nlambda_h
*     integer
*     Number of wavelengths lambda_h.
**nlambda_t
*     integer
*     Number of wavelengths lambda_t.
**nlev
*     integer
*     Number of atmospheric levels.
**rad
*     real array(levels, lambda_t)
*     The radiation field at resolution lambda_t and nlev levels.
*OUTPUT
**rad_h
*     real array(levels,lambda_h)
*     The radiative quantity at lambda_h (high) resolution
*ERROR_MESSAGES
*ROUTINES_CALLED
*     zspl3
*     spl3
*FILES
*     param.inc                Include file
*REFERENCES
*EOD
*
      IMPLICIT LOGICAL (A-Z)    ! To nearly get strong typing
      INCLUDE 'param.inc'
*
      INTEGER iv, lu, nlambda_h, nlambda_t, nlev, maxlev
      REAL lambda_h(*), lambda_t(*), rad(maxlev,*), rad_h(maxlev,*)
*
* Qubic spline variables
*
      REAL bspl3(maxwvn), splcof(maxwvn), dummy(maxwvn),
     $     uspl3(maxwvn), vspl3(maxwvn), wspl3(maxwvn)
      REAL spl3
*
      DO lu = 1, nlev
         DO iv = 1, nlambda_t
            dummy(iv) = rad(lu,iv)
         ENDDO
*
*        Qubic spline coefficients
         CALL zspl3( nlambda_t,lambda_t,dummy,
     $        wspl3, bspl3, uspl3, vspl3, splcof )
*
         DO iv = 1, nlambda_h
            rad_h(lu,iv) = spl3( nlambda_t, lambda_t,
     $           dummy, splcof, lambda_h(iv) )
         ENDDO
      ENDDO
      RETURN
      END

