/* lex.l                                                -*- C -*-

Copyright (C) 1995 Arve Kylling

This file is part of libRadtran.

Radtran is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Radtran is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

%{

#include <getopt.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "uvspec.h"
#include "uvspec_anc.h"

float  yytext2float(char* yytext, int which_token, int max_token);
float* yytext2floats(char* yytext, int nary);
int    yytext2nfloats(char* yytext);
int    yytext2int(char* yytext, int which_token, int max_token);
int*   yytext2integers(char* yytext, int nary);
int    yytext2nintegers(char* yytext);
char*  yytext2string(char* yytext, int which_token, int max_token);

char* s;
float* tmp;
int ierror=0;
int start_condition = INITIAL;
int line_number;
input_struct Input;
output_struct Output;

%}

%option noyywrap

DIGIT    [0-9]
SPACE    [ \t]
CCHAR    [#]
COMMENT  ({CCHAR}.*)
INTEGER  {DIGIT}+
SPACE_INTEGER {SPACE}+{INTEGER}
INTEGERS {SPACE_INTEGER}+ 
EXPON   ([DdEe][+-]?{DIGIT}+)
FLOAT    ((({DIGIT}+"."{DIGIT}*)|({DIGIT}+"."{DIGIT}*{EXPON}))|({INTEGER}))
SPACE_FLOAT {SPACE}+{FLOAT}
FLOATS   {SPACE_FLOAT}+ 
SIGN     [-+]
SIGNED_FLOAT {SIGN}?{FLOAT}
SPACE_SIGNED_FLOAT {SPACE}+{SIGNED_FLOAT}
SIGNED_FLOATS {SPACE_SIGNED_FLOAT}+
SIGNED_INTEGER {SIGN}?{INTEGER}
WORD     [_\.\/A-Za-z][_\.\/A-Za-z0-9\-]*

%%

  /* These are the allowed input lines in the uvspec input file 
     The input line specifications must come before the eating of
     blanks, otherwise things get messed up 
     Obviously this whole lot could have been handled more properly
     using bison etc., but than rather few would know what was going 
     on??? 
   */

aerosol_haze{SPACE}+{INTEGER}       {Input.aer.haze     = yytext2int(yytext,2,2); }
aerosol_scale_ssa{SPACE}+{FLOAT}    {Input.aer.scale_ssa= yytext2float(yytext,2,2); }
aerosol_set_gg{SPACE}+{FLOAT}       {Input.aer.set_gg   = yytext2float(yytext,2,2); }
aerosol_set_ssa{SPACE}+{FLOAT}      {Input.aer.set_ssa  = yytext2float(yytext,2,2); }
aerosol_season{SPACE}+{INTEGER}     {Input.aer.seasn    = yytext2int(yytext,2,2); }
aerosol_tau_file{SPACE}+{WORD}      {Input.aer.tau_filename
				       = yytext2string(yytext,2,2); }
aerosol_visibility{SPACE}+{FLOAT}   {Input.aer.visib    = yytext2float(yytext,2,2); }
aerosol_vulcan{SPACE}+{INTEGER}     {Input.aer.vulcan   = yytext2int(yytext,2,2); }
albedo{SPACE}+{FLOAT}               {Input.atm.albedo   = yytext2float(yytext,2,2); }
albedo_file{SPACE}+{WORD}           {Input.albedo_filename  = yytext2string(yytext,2,2);
                                     Input.atm.albedo   = -1.;}
altitude{SPACE}+{SIGNED_FLOAT}      {Input.atm.altitude = yytext2float(yytext,2,2); }
angstrom{SIGNED_FLOATS}             {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.aer.alpha    = yytext2float(s,2,3); 
                                     Input.aer.beta     = yytext2float(yytext,3,3);
				     free(s);}
atmosphere_file{SPACE}+{WORD}       {Input.atmosphere_filename 
				       = yytext2string(yytext,2,2); }
crs_file{SPACE}+{INTEGER}{SPACE}+{WORD} {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
                                     Input.crs.column   = yytext2int(s,2,3);
                                     Output.atm.ndens    = Input.crs.column - 3 + 1;
                                     /* -3 because minus z, p and T column */
                                     Input.crs.filename 
				       = yytext2string(yytext,3,3); }
data_files_path{SPACE}+{WORD}       {Input.filepath     = yytext2string(yytext,2,2); }
day_of_year{SPACE}+{INTEGER}        {Input.day_of_year  = yytext2int(yytext,2,2); }
deltam{SPACE}+{WORD}                {s =  yytext2string(yytext,2,2);
				     if (strncmp("on",s,2)==0) {Input.rte.deltam = 1;}
				     else if (strncmp(s,"off",3)==0) {Input.rte.deltam = 0;}
				     free(s);}
fisot                               {Input.rte.fisot    = 1; }
flexstor                            {Input.flexstor     = 1; }
header                              {Input.header       = 1; }
nstr{SPACE}+{INTEGER}               {Input.rte.nstr     = yytext2int(yytext,2,2); }
o3_crs{SPACE}+{WORD}                {s =  yytext2string(yytext,2,2);
				     if (strncmp("Bass_and_Paur",s,13)==0) {Input.crs.o3_type = O3_BASSPAUR;}
				     else if (strncmp(s,"Molina",6)==0)    {Input.crs.o3_type = O3_MOLINA;}
				     else if (strncmp(s,"Daumont",7)==0)   {Input.crs.o3_type = O3_DAUMONT;}
				     else {fprintf(stderr,"Unknown 'o3_crs %s' on line %d\n", s, line_number+1);
				     ierror++;}
				     free(s);}
ozone_column{SPACE}+{SIGNED_FLOAT}  {Input.ozone_column = yytext2float(yytext,2,2); }
phi0{SPACE}+{SIGNED_FLOAT}          {Input.atm.phi0     = yytext2float(yytext,2,2); }
pressure{SPACE}+{SIGNED_FLOAT}      {Input.pressure = yytext2float(yytext,2,2); }
prndis{INTEGERS}                    {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.rte.nprndis  = yytext2nintegers(s); 
				     Input.rte.prndis   = 
				       yytext2integers(yytext, Input.rte.nprndis); 
				     free(s); }
reverse                             {Input.rte.reverse    = 1; }
rte_solver{SPACE}+{WORD}            {s =  yytext2string(yytext,2,2);
				     if (strncmp("sdisort",s,7)==0) {Input.rte.solver = SDISORT;}
				     else if (strncmp(s,"disort",6)==0) {Input.rte.solver = DISORT;}
				     else if (strncmp(s,"sucordsca",9)==0) {Input.rte.solver = SUCORD;}
				     else if (strncmp(s,"twostr",6)==0) {Input.rte.solver = TWOSTR;}
				     else if (strncmp(s,"montecarlo",10)==0) {Input.rte.solver = MONTECARLO;}
				     else {fprintf(stderr,"Unknown 'rte_solver %s' on line %d\n", s, line_number+1);
				     ierror++;}
				     free(s);}
slit_function_file{SPACE}+{WORD}    {Input.slit_function_filename 
				       = yytext2string(yytext,2,2);
                                     Input.convolve = 1; }
spline{SIGNED_FLOATS}               {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.spline_lambda_0    = yytext2float(s,2,4); 
				     strcpy (s, yytext);
				     Input.spline_lambda_1    = yytext2float(s,3,4); 
                                     Input.spline_lambda_step = yytext2float(yytext,4,4);
                                     Input.spline = 1;
				     free(s);}
spline_file{SPACE}+{WORD}           {Input.spline_filename 
				       = yytext2string(yytext,2,2);
                                     Input.spline = 1; }
solar_file{SPACE}+{WORD}            {Input.extraterrestrial_filename 
				       = yytext2string(yytext,2,2); }
sza{SPACE}+{SIGNED_FLOAT}           {Input.atm.sza      = yytext2float(yytext,2,2); }
sza_file{SPACE}+{WORD}              {Input.sza_filename = yytext2string(yytext,2,2);
                                     Input.atm.sza      = -1000.0;}
transmittance_wl_file{SPACE}+{WORD} {Input.transmittance_wl_file 
				       = yytext2string(yytext,2,2); }
umu{SIGNED_FLOATS}                  {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.rte.numu     = yytext2nfloats(s); 
				     Input.rte.maxumu   = Input.rte.numu;
				     Input.rte.umu      = 
				       yytext2floats(yytext, Input.rte.numu); 
				     free(s); }
phi{SIGNED_FLOATS}                  {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.rte.nphi     = yytext2nfloats(s); 
				     Input.rte.maxphi   = Input.rte.nphi;
				     Input.rte.phi      = 
				       yytext2floats(yytext, Input.rte.nphi); 
				     free(s); }
wc_file{SPACE}+{WORD}               {Input.watercloud_filename 
				       = yytext2string(yytext,2,2); }
wc_set_tau{SPACE}+{FLOAT}           {Input.wc.wctau  = yytext2float(yytext,2,2); }
wvn{FLOATS}                         {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.wl.wl_start  = yytext2float(s,2,3); 
                                     Input.wl.wl_end    = yytext2float(yytext,3,3);
				     free(s);}
zout{FLOATS}                        {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
				     strcpy (s, yytext);
				     Input.atm.nzout    = yytext2nfloats(s); 
                                     Input.atm.zout     = 
				       yytext2floats(yytext, Input.atm.nzout); 
				     free(s); }


{SPACE}+       ;         /* eat up blanks */ 

{COMMENT}*     ;         /* eat up comments */

\n                        { ++line_number;};

{WORD}                    {fprintf(stderr,"Unknown command on line %d\n", line_number+1);
                           fprintf(stderr,"%s\n", yytext);
                           ierror++;}

<<EOF>> { 
  if ( ierror > 0 ) { exit(1); }
	    yyterminate();
        }      

%%

/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void usage()
{

  fprintf (stderr, "\nuvspec, version %4.2f\n", VERSION);
  fprintf (stderr, "\nCalculates the 1D radiation field of the Earth\'s atmosphere.\n");
  fprintf (stderr, "See libRadtran documentation for more information.\n\n");
  fprintf (stderr, "No error checking yet, so be careful.\n\n");

  fprintf (stderr, "Usage: uvspec [-h] < input_file > Output_file\n");
  fprintf (stderr, " -h    Print this message.\n");  
  fprintf (stderr, "\n");  
}

int main(int argc, char **argv) 
{
  
  /*  @50c@                

@code{uvspec} calculates the radiation field in the Earth's atmosphere
for a given set of input parameters. It reads input from standard input, 
and outputs to standard output. It is normally invoked in the following
way: 

@example
uvspec < input_file > output_file
@end example

@noindent The format of the input and output files are described
below. Several realistic examples of input files are subsequently 
given.

@strong{Warning:} Please note the error checking on input variables 
is very scarce at the moment. Hence, if you provide erroneous input, 
the outcome is unpredictable.
  @c50@ */

  /*  @50_01c@                
@cindex input file, uvspec
@cindex uvspec, input file

The uvspec input file consists of single line entries, each making
up a complete input to the uvspec program. First on the line comes
the parameter name, followed by one or more parameter values. The 
parameter name and the parameter values are seperated by white space.

Filenames are entered without any surrounding single or double quotes.

Comments are introduced by a @code{#}. Blank lines are ignored.

The various input parameters are described in detail below.
@vtable @code
@item aerosol_haze
Specify aerosol type in the lower 2 km of the atmosphere. Integer.
@table @code
@item 1
Rural type aerosols.
@item 4
Maritime type aerosols.
@item 5
Urban type aerosols.
@item 6
Tropospheric type aerosols.
@end table
@item aerosol_scale_ssa
Scale the aerosol single scattering albedo for all wavelengths and altitudes
with number, must be larger than 0.0. If the resulting scaled single scattering
albedo is larger than 1.0 it is set to 1.0.
@item aerosol_set_gg
Set the aerosol asymmetry factor for all wavelengths and altitudes
to a float between -1.0 and 1.0.
@item aerosol_set_ssa
Set the aerosol single scattering albedo for all wavelengths and altitudes
to a float between 0.0 and 1.0.
@item aerosol_season
Specify season to get appropriate aerosol profile.
@table @code
@item 1
Spring-summer profile.
@item 2
Fall-winter profile.
@end table
@item aerosol_tau_file
Location of aerosol optical depth file. The file must have two columns.
Column 1 is the altitude in km. The altitude grid must be exactly equal to 
the altitude grid specified in the file @code{atmosphere_filename}.
Column 2 is the aerosol optical depth. To get wavelength dependence use
the parameter @code{angstrom}. Comment lines in the file are introduced 
by a @code{#}.
@item aerosol_visibility
Visibility in km.
@item aerosol_vulcan
Specify aerosol situation above 2 km. Integer.
@table @code
@item 1
Background aerosols.
@item 2
Moderate vulcanic aerosols.
@item 3
High vulcanic aerosols.
@item 4
Extreme vulcanic aerosols.
@end table
@item albedo
The Lambertian surface albedo. A float number between 0.0 and 1.0. 
The same for all wavelengths. For wavelength dependent surface albedo 
use @code{albedo_file}.
@item albedo_file
Location of surface albedo file for wavelength dependent surface albedo. 
The file must have two columns.
Column 1 is the wavelength, in nm, and column 2 the corresponding 
Lambertian surface albedo. The wavelength grid may be freely set. The 
albedo will be interpolated to the wavelength grid used for the radiation
calculation. Comment lines in the file are introduced by a @code{#}.
@item angstrom
Scale the aerosol optical depth using the @AA{}ngstr@"om formula. Specify 
the @AA{}ngstr@"om alpha and beta coefficients. @code{aerosol_vulcan}, 
@code{aerosol_season}, @code{aerosol_haze} and @code{aerosol_visibility} must also
be specified.
@item altitude
Set the bottom level in the model atmosphere provided in 
@code{atmosphere_filename} to be at altitude (km).
@example
altitude 0.73   # Altitude of IFU, Garmisch-Partenkirchen 
                # Beaware, for this to work the atmosphere file 
                # must start at 0 km.
@end example
The profiles are cut at the specified altitude. This is a convenient way for 
the user to calculate the irradiance at other locations at other
altitudes than sealevel. Note that @code{altitude} is very different
from @code{zout}. Also note that @code{zout} is 0 km at @code{altitude}.
@item atmosphere_filename
Location of the atmospheric data file. The first five columns of the file are
read and must come in the following order:
@example
z(km) p(mb) T(K) air_density(cm-3) o3_density(cm-3)
@end example
where @code{z} is the altitude above sea level. The atmosphere is specified
top-down, that is, the top level is the first line in the file, the bottom
(surface) level the last line. It is important that the correct units are 
used. Otherwise unpredictable results are guaranteed.
Comment lines in the file are introduced by a @code{#}.
@item data_files_path
The path to the directory where all uvspec internal data files lives, e.g.
the files that are in the subdirectories of the @code{data} directory
that comes with the uvspec distribution.
@item day_of_year
If given the calculated radiative quantities will be corrected
for the Earth-Sun distance. If not specified the Earth-Sun distance is
1 AU (i.e. equinox distances). @code{day_of_year} must be an integer.
@item deltam
Turn delta-M scaling on/off. Set to either @code{on} or @code{off}.
@item fisot
Specifies that isotropic illumination is used at top-boundary instead
of beam source. Useful for those who want to calculate the reflectance 
for a homogeneous or inhomogeneous atmosphere. The intensity is still set by
@code{solar_file}.
The Lambertian surface albedo. A float number between 0.0 and 1.0. 
The same for all wavelengths. For wavelength dependent surface albedo 
use @code{albedo_file}.
@item flexstor
Output is in flexstor format. May not be combined with @code{header}.
Also, does not currently work when @code{umu} and/or @code{phi} is specified.
@item header
Output information on some of the input parameters. May not be combined
with @code{flexstor}.
@item nstr
Number of streams used to solve the radiative transfer equation.
Default is 6. If changed, make sure that @code{mxcmu} in 
@file{src_f/disort.f} and @file{src_f/sdisort.f} are larger or equal 
to @code{nstr}. If not, change @code{mxcmu} and recompile.
@item o3_crs
Choose between the following ozone cross sections.
@table @code
@item Bass_and_Paur
Bass and Paur ozone cross section.
@item Molina
Molina and Molina ozone cross section.
@item Daumont
Daumont et al. ozone cross section.
@end table
Molina and Molina is default.
@item ozone_column
Scale the ozone column to a given value. The value must be in 
Dobson units. If value is negative or @code{ozone_column}
is not included, the ozone column is not scaled.
@item phi
Azimuth output angles (in degrees) in increasing order. 
The radiance is output at @code{phi} and @code{umu}.
@item phi0
Azimuth angles of the sun (0 to 360 degrees). If @code{phi0} varies as a
function of wavelength use @code{sza_file}.
@item pressure
The surface pressure in mb. The pressure profile and density profiles are 
scaled accordingly.
@item prndis
Specify one or more integers between 1 and 7.
Print various disort input and output in disorts own format. See 
@file{disort.doc} for more information.
@strong{Warning:} Produces a lot of output. 
@item reverse
Option for the strong and bold. Reverses the atmospheric input to the radiative
transfer solvers. That is, the atmosphere is turned on the head.
Yes, that is actually useful for some purposes. If you think
you need this contact the author. Otherwise, do not use.
@item rte_solver
Set the radiative transfer equation solver to be used. Options are
@vtable @code
@item disort
The standard plane--parallel disort algorithm by Stamnes et al. (1988).
@item sdisort
Pseudospherical disort as described by Dahlback and Stamnes (1991).

@strong{Warning:} it is not recommended to use sdisort for cloudy conditions. 
With clouds included it may be numerically unstable and produce 
wrong results.
@item twostr
The two--stream radiative transfer solver described by Kylling, Stamnes
and Tsay (1995).
@end vtable
@item slit_function_file
If specified, the calculated spectrum is convolved with the function found
in the @file{slit_function_file}. The file must
contain two columns. Column 1 is the wavelength, in nm, and relative to the
center wavelength. Column 2 is the corresponding slit function value. It must
be unity at the maximum. The wavelength steps in the slit function file must
be equidistant. Comment lines in the file are introduced by a @code{#}.
@item solar_file
Location of file holding the extraterrestrial spectrum. The file must
contain two columns. Column 1 is the wavelength, in nm, and column 2 
the corresponding extraterrestrial flux. The user may freely use any units
he/she wants on the extraterrestrial flux. The wavelength grid specified, 
will be the wavelength resolution at which results are returned. However, 
the wavelength range is determined by @code{wvn}.
Comment lines in the file are introduced by a @code{#}.
@item spline
Spline interpolate to wavelengths lambda_0 to lambda_1 in steps of lambda_step,
in nm. Specified as e.g.
@example
spline 290. 365. 0.5
@end example
Here, the calculated spectrum is interpolated to wavelengths 290., 290.5, 291., ...
, 364.5, 365. For interpolation to arbitrary wavelengths use @code{spline_file}.
The specified wavelength interval must be within the one specified by @code{wvn}.
@item spline_file
Spline interpolate to arbitrary wavelengths, in nm, given as a single column in file 
@file{splin_file}. The specified wavelengths must be within the one specified 
by @code{wvn}. Comment lines in the file are introduced by a @code{#}.
@item sza
The solar zenith angle. If the solar zenith angle varies with wavelength,
use @code{sza_file}. 
@item sza_file
Location of solar zenith angle file for wavelength dependent solar
zenith angle. The file must have two or three columns. Column 1 is the wavelength, 
in nm, and column 2 the corresponding  solar zenith angle. Optinally
the third column may contain the corresponding solar azimuth angle. The solar 
azimuth angle is only needed when calculating radiances. The wavelength 
grid may be freely set. The solar zenith and azimuth angle will be interpolated to 
the wavelength grid used for the radiation calculation. Comment lines 
in the file are introduced by a @code{#}.
@item transmittance_wl_file
Location of single column file that sets the wavelength grid used for the 
internal transmittance calculations. The wavelengths must be in nm.
Do not use this option unless you know what you are doing.
@item umu
Cosine of output polar angles in increasing order, starting with 
negative (downward) values (if any) and on through positive (upward)
values. Must not have any zero values. The azimuthally averaged intensity
@code{u0u} is output at @code{umu}.
@item wc_file
Location of file specifying optional water cloud properties. The file must
contain three columns. Column 1 is the altitude, in km, column 2 
the liquid water content in grams per qubic meter, and column 3 the effective
droplet radius in micron. The liquid water content and effectice droplet 
radius must be specified at the same levels, i.e. altitude grid, as specified
in the @code{atmosphere_file}. Note that the definition of cloud altitudes   
in @code{wc_file} refers to sea level, not to altitude above
ground. E.g., when altitude is set to 1.63km, and the first cloud level is 
defined at 3km, the cloud would start at 1.37km above ground.
The optical depth of a layer is calculated 
using information from the upper and lower levels defining the layer. An example 
of a cloud defined by three levels and thus covering two model layers is given in 
the file @file{examples/WC.DAT}.   
The parameterization due to Hu and Stamnes (An accurate parameterization of the 
radiative properties of water clouds suitable for use in climate models, J. Clim., 
6, 728-742, 1993). Comment lines in the file are introduced by a @code{#}.
@item wc_set_tau
Set the total water cloud optical depth for all wavelengths.
@item wvn
Sets the wavelength range, in nm. Specify first wavelength and last 
wavelength. The wavelength resolution is specfied in the @code{solar_file}.
@item zout 
Output altitudes in km, one or more altitudes may be specified in increasing
magnitude. Output altitudes must be within the atmosphere specified in the
@code{atmosphere_file}. Note that @code{zout} does not restructure your 
atmosphere model. Hence, if you specify @code{zout 0.730} and have your
atmosphere model in @code{atmosphere_filename} go all the way down to 
sea level, i.e. 0.0km., output is presented at 0.730km and calculations
performed with an atmosphere between 0.0 and 0.730 km (and above of course). 
If you want calculations done for e.g. a mountain site you have to restructure
the atmosphere model and make sure it stops at the appropriate altitude.
This you may either due by editing the atmosphere file or by using 
@code{altitude}.
@end vtable
  @c50_01@ */

  /*  @50_02c@                
@cindex output, uvspec
@cindex uvspec, output
The uvspec output is currently fixed. We hope to make it user controllable
in the future.

Uvspec outputs one line to standard output (stdout) for each wavelength.
The format of the output line for the disort and sdisort solvers is

@example
lambda rfldir rfldn flup uavgso uavgdn uavgup
@end example

@noindent if @code{umu} is not specified, and

@example
lambda rfldir rfldn flup uavgso uavgdn uavgup (u0u(iu),iu=1,numu)
@end example

@noindent if @code{umu} is specified. If both @code{umu} and @code{phi} is specified
the output format is

@example
lambda rfldir rfldn flup uavgso uavgdn uavgup
                            phi(0)     ...        phi(m)
umu(0) u0u(umu(0))   uu(umu(0),phi(0)) ... uu(umu(0),phi(m))
umu(1) u0u(umu(1))   uu(umu(1),phi(0)) ... uu(umu(1),phi(m))
  .     .             .                     .
  .     .             .                     .
umu(n) u0u(umu(n))   uu(umu(n),phi(0)) ... uu(umu(n),phi(m))
@end example

@noindent and so on for each wavelength.

The format of the output line for the twostr solver is

@example
lambda rfldir rfldn flup uavg
@end example

Here

@vtable @code
@item lambda
Wavelength (nm)
@item rfldir
Direct beam irradiance (same units as extraterrestrial irradiance, e.g mW/(m2 nm) if 
using the @file{atlas3} spectrum in the @file{data/solar_flux} directory.)
@item rfldn
Diffuse down irradiance, i.e. total minus direct beam (same units
as @code{rfldir}).
@item flup
Diffuse up irradiance (same units as @code{rfldir}).
@item uavg
The mean intensity. Proportional to the actinic flux. (same units as @code{rfldir}).
@item uavgso
Direct beam contribution to the mean intensity. (same units as @code{rfldir}).
@item uavgdn
Diffuse downward radiation contribution to the mean intensity. (same units as @code{rfldir}).
@item uavgup
Diffuse upward radiation contribution to the mean intensity. (same units as @code{rfldir}).
@item u0u
The azimuthally averaged intensity at @code{numu} user specified 
angles @code{umu}. (units of e.g. mW/(m2 nm sr) if using the @file{atlas3} spectrum 
in the @file{data/solar_flux} directory.)  
@item uu
The radiance (intensity) at @code{umu} and @code{phi} user specified 
angles. (units of e.g. mW/(m2 nm sr) if using the @file{atlas3} spectrum 
in the @file{data/solar_flux} directory.)  

@end vtable

@noindent The total downward irradiance is given by

@example
irr_down = rfldir + rfldn
@end example

@noindent The total mean intensity is given by

@example
uavg = uavgso + uavgdn + uavgup
@end example

@noindent If @code{deltam} is on it does not make sense to look at the individual 
contributions to @code{uavg} since they are delta-M scaled.

  @c50_02@ */

  /*  @50_03c@                

An example of a complete input file for a clear sky atmosphere is

@example
@include ../examples/UVSPEC_CLEAR.INP
@end example

@noindent 
The atmosphere model, i.e. pressure, temperature, and ozone concentrations 
profiles are read from  the file @code{../data/atmmod/afglus.dat}. The 
extraterrestrial solar flux is read from the file 
@code{../data/solar_flux/atlas_plus_modtran}.

A wavelength dependent surface albedo may be specified by using @code{albedo_file}
instead of @code{albedo}.

For simulations of measured spectra the change in solar zenith angle during a 
scan must be accounted for. This is done by using @code{sza_file} instead of
@code{sza}. 

Aerosol information is usually of importance when simulating measured spectra.
To include aerosols the following parameters must be set: @code{aerosol_vulcan},
@code{aerosol_season}, @code{aerosol_haze} and @code{aerosol_visibility}. Using these
parameters will use the aerosol models of Shettle (1989) to calculate the aerosol
optical depth, single scattering albedo and asymmetry factor. The aerosol optical
depth may be scaled by using the @code{angstrom} parameter. If information about the
aerosol optical depth profile is available it may be specified in @code{aerosol_tau_file}.
The aerosol single scattering albedo may be scaled by @code{aerosol_scale_ssa} and 
set by @code{aerosol_set_ssa}.  The aerosol asymmetry factor may be set by
@code{aerosol_set_gg}. Instead of using the standard aerosol profiles from the 
aerosol models an aerosol optical depth profile may be specified using 
@code{aerosol_tau_file}. The wavelength dependence of the aerosol optical depth is
specified using the @code{angstrom} parameter.

A rather complete example is

@example
@include ../examples/UVSPEC_AEROSOL.INP
@end example

@noindent 
By combining this with the clear sky example given above a complete uvspec
input file including aerosol is constructed.

To calculate radiances the following lines will do the job when combined
with the e.g. the clear sky example above

@example
@include ../examples/UVSPEC_RADIANCES.INP
@end example

Water clouds may be included by specifying @code{wc_file}.

@example
@include ../examples/UVSPEC_WC.INP
@end example

@noindent 
The file @code{WC.DAT} is found in the examples directory.

More examples with output are found in the @file{examples} directory.

      @c50_03@   */

  /***************** End of documentation *********************/

  int c;
  int iu, iv, j, lev;
  int status=0;
  /* Input_struct Input;
  Output_struct Output; */

  /* Initialization */

  Input.flexstor      = 0;
  Input.header        = 0;
  Output.print.phi    = 1; /* Default print radiances */
  Output.atm.ndens    = 2; /* Default value for atmospheric files with
                              air and o3 densities. */
  Input.slit_function_filename = "";
  Input.convolve      = 0;
  Input.spline      = 0;
  Input.spline_filename = "";
  Input.albedo_filename = "";
  Input.atmosphere_filename = "";
  Input.crs.o3_type = O3_MOLINA;
  Input.crs.filename = "";
  Input.day_of_year     = -999;
  Input.extraterrestrial_filename = "";
  Input.filepath        = "";
  Input.ozone_column    = -999;
  Input.pressure        = -999;
  Input.transmittance_wl_file = "";
  Input.watercloud_filename       = "";
  Input.aer.alpha       = 0.0;
  Input.aer.beta        = -999.;
  Input.aer.haze        = 0;
  Input.aer.scale_ssa   = 0;
  Input.aer.set_gg      = -999.;
  Input.aer.set_ssa     = -1.;
  Input.aer.seasn       = 0;
  Input.aer.tau_filename= "";
  Input.aer.visib       = 50.;
  Input.aer.vulcan      = 0;
  Input.atm.albedo      = 0.0;
  Input.atm.altitude    = 0.0;
  Input.atm.sza         = 0.0;
  Input.rte.fisot       = 0;
  Input.rte.maxphi      = 0;
  Input.rte.nstr        = 6;
  Input.rte.solver      = DISORT;
  Input.wl.wl_start     = 0.0;
  Input.wl.wl_end       = 0.0;
  Input.rte.nphi        = 0;
  Input.rte.numu        = 0;
  Input.rte.nprndis     = 0;
  Input.rte.deltam      = 1;
  Input.rte.reverse     = 0;
  Input.atm.nzout       = 0;
  Input.wc.wctau        = -999;

  while ((c=getopt (argc, argv, "h")) != EOF)  {
    switch(c)  {
    case 'h': 
      usage();
      return (-1);
      break;
    default:
      usage();
      return (-1);
    }
  }

  /* Read the input file */

  yylex();

  if (Input.rte.nphi <= 0) {
    Input.rte.maxphi      = 3; /* Minimum number required by disort. */
    Input.rte.nphi = 1;
    Input.rte.phi = (float *) calloc (Input.rte.maxphi, sizeof(float));
    Output.print.phi    = 0;  /* Do not print radiances if user do not want them */
  }
  if (Input.rte.numu <= 0) {
    Input.rte.maxumu      = Input.rte.nstr;
    Input.rte.umu = (float *) calloc (Input.rte.maxumu, sizeof(float));
  }

  if (Input.atm.nzout == 0) {
    Input.atm.nzout = 1;
    Input.atm.zout = (float *) calloc (Input.atm.nzout, sizeof(float));
    Input.atm.zout[0] = 0.0;
  }

  status = uvspec_check(Input);
  if (status!=0) {
    fprintf (stderr, "%d error(s) in uvspec input-file, aborting\n", status);
    return status;
  }

  /* Do the radiation calculation */

  status = uvspec(Input,&Output);
  if ( status != 0.0 ) {
    fprintf (stderr, "Error during execution of uvspec\n");
    return status;
  }

  /* Convolve with slitfunction */
  if ( Input.convolve ) {
    status = convolve(Input,&Output);
    if ( status != 0.0 ) {
      fprintf (stderr, "Error during execution of uvspec: convolve\n");
      return status;
    }
  }

  /* Spline interpolate to output grid */
  if ( Input.spline ) {
    status = spline_interpolate(Input,&Output);
    if ( status != 0.0 ) {
      fprintf (stderr, "Error during execution of uvspec: spline\n");
      return status;
    }
  }

 

  /* Print output ***FIXME*** THIS SHOULD BE USER DEFINED SOMEHOW */

  /*    printf("Ecc: %f\n",Output.eccentricity_factor); */
  /*  printf("nlambda_t: %d %d, nzout: %d\n",Output.wl.nlambda_t, Output.wl.nlambda_h, Input.atm.nzout);  */
  /*      printf("Debug %d \n", sizeof(Output.u0u));*/
  if (Input.flexstor) {
    printf("flexstor 1 0 suvdama 1 1\n");
    printf("# This flexstor output has not been checked by flxcheck.\n");
    printf("# Hence it may not follow flexstor/suvdama rules. This will\n");
    printf("# be improved soon.\n");
    printf("model_name() uvspec, version %4.2f\n", VERSION);
    if ( fabs(Input.aer.beta-0.0) > 0.0 ) {
      printf("angstrom_alpha()  %f\n", Input.aer.alpha);
      printf("angstrom_beta()  %f\n", Input.aer.beta);
    }
    printf("ozone_column(DU) %f\n", Output.ozone_column);
    printf("pressure(mb) %f\n", Input.pressure);
    printf("self_day_of_year() %d\n", Input.day_of_year);
    /*    printf("self_zout(km)");
    for (lev=0;lev<Input.atm.nzout;lev++) {
      printf("  %f", Input.atm.zout[lev]);
    }      
    printf("\n");*/
    printf("self_temper(K)");
    for (lev=0;lev<Input.atm.nzout;lev++) {
      printf("  %f", Output.atm.temper_out[lev]);
    }      
    printf("\n");
    printf("self_day_of_year() %d\n", Input.day_of_year);
    printf("self_eccentricity_factor() %f\n", Output.eccentricity_factor);
    printf("begin table data\n");
    printf("WL(nm) SZA(deg.dec) self_rfldir(mW m-2 nm-2) self_rfldn(mW m-2 nm-2) self_rflup(mW m-2 nm-2)");
    printf(" self_uavgso(mW m-2 nm-2) self_uavgdn(mW m-2 nm-2) self_uavgup(mW m-2 nm-2)\n");
    for (iv=0;iv<Output.wl.nlambda_h;iv++) {
      for (lev=0;lev<Input.atm.nzout;lev++) {
	printf("%7.3f %7.3f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e", 
	       Output.wl.lambda_h[iv], Output.sza[iv],
	       Output.rfldir[iv][lev], Output.rfldn[iv][lev], Output.flup[iv][lev],
	       Output.uavgso[iv][lev], Output.uavgdn[iv][lev], Output.uavgup[iv][lev]);
      }
      printf("\n");
    }
    printf("end table data\n");
  }
  else {
    if (Input.header) {
      printf("#aerosol_haze                 : %16d\n", Input.aer.haze);
      printf("#aerosol_season               : %16d\n", Input.aer.seasn);
      printf("#aerosol_set_gg               : %16.3f\n", Input.aer.set_gg);
      printf("#aerosol_set_ssa              : %16.3f\n", Input.aer.set_ssa);
      printf("#aerosol_tau_file             : %s\n", Input.aer.tau_filename);
      printf("#aerosol_visibility           : %16.3f\n", Input.aer.visib);
      printf("#albedo                       : %16.3f\n", Input.atm.albedo);
      printf("#angstrom                     : %16.3f  %16.3f\n", Input.aer.alpha, Input.aer.beta);
      printf("#atmosphere_file              : %s\n", Input.atmosphere_filename);
      printf("#day_of_year                  : %16d\n", Input.day_of_year);
      switch (Input.rte.deltam) {
      case 0:
	printf("#deltam                       : off\n");
	break;
      case 1:
	printf("#deltam                       : on\n");
	break;
      }
      switch (Input.crs.o3_type) {
      case O3_BASSPAUR:
	printf("#o3_crs                       : Bass_and_Paur\n");
	break;
      case O3_MOLINA:
	printf("#o3_crs                       : Molina\n");
	break;
      case O3_DAUMONT:
	printf("#o3_crs                       : Daumont\n");
	break;
      }
      printf("#ozone_column                 : %16.3f\n", Input.ozone_column);
      switch (Input.rte.solver) {
      case SDISORT:
	printf("#rte_solver                   : sdisort\n");
	break;
      case DISORT:
	printf("#rte_solver                   : disort\n");
	break;
      case SUCORD:
	printf("#rte_solver                   : sucorcsca\n");
	break;
      case TWOSTR: 
	printf("#rte_solver                   : twostr\n");
	break;
      }
      printf("#slit_function_file           : %s\n", Input.slit_function_filename);
      printf("#solar_file                   : %s\n", Input.extraterrestrial_filename);
      printf("#wvn                          : %16.3f  %16.3f\n", Input.wl.wl_start, Input.wl.wl_end);
      printf("#zout                         : ");
      for (lev=0;lev<Input.atm.nzout;lev++) {
	printf(" %16.3f", Input.atm.zout[lev]);
      }
      printf("\n");
    }
    for (iv=0;iv<Output.wl.nlambda_h;iv++) {
      for (lev=0;lev<Input.atm.nzout;lev++) {
	if (Input.rte.solver == TWOSTR) {
	  printf("%7.3f %13.6e %13.6e %13.6e %13.6e ", 
		 Output.wl.lambda_h[iv], 
		 Output.rfldir[iv][lev], Output.rfldn[iv][lev], Output.flup[iv][lev],
		 Output.uavg[iv][lev]);
	}
	else {
	  printf("%7.3f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e", 
		 Output.wl.lambda_h[iv], 
		 Output.rfldir[iv][lev], Output.rfldn[iv][lev], Output.flup[iv][lev],
		 Output.uavgso[iv][lev], Output.uavgdn[iv][lev], Output.uavgup[iv][lev]);
	}
	if ( Input.rte.numu > 0 && !Output.print.phi ) {
	  for (iu=0;iu<Input.rte.numu;iu++) {
	    printf(" %13.6e ", Output.u0u[iv][lev][iu]);
	  }
	} 
	printf("\n");
	if ( Output.print.phi ) {
	  printf("                      ");
	  for (j=0;j<Input.rte.nphi;j++) {
	    printf(" %13.3f ", Input.rte.phi[j]);
	  }
	  printf("\n");
	  for (iu=0;iu<Input.rte.numu;iu++) {
	    printf("%7.4f %13.6e ", Input.rte.umu[iu], Output.u0u[iv][lev][iu]);
	    for (j=0;j<Input.rte.nphi;j++) {
	      printf(" %13.6e ", Output.uu[iv][lev][j][iu]);
	    }
	    printf("\n");
	  }
	} 
      }
    }    
  }
  return status;

}

float yytext2float(char* yytext, int which_token, int max_token) {
  double dnumber;
  char token[] = " ";
  char *p, *s=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments on line %d\n", line_number+1);
  }
  dnumber = atof(s);
  /*   printf("dnumner %f\n",dnumber);*/
  return (float)dnumber;
}

float* yytext2floats(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;

  /* Fill up ary */
  float *ary    = (float *) calloc (nary, sizeof(float));

  p = strtok(yytext,token);
  i = 0;
  while ( (p = strtok(NULL,token)) != NULL) {
    ary[i++] = atof(p);
  }
  return ary;
}

int yytext2nfloats(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext,token);
  while ( (p = strtok(NULL,token)) != NULL) {
    nary++;
  }
  return nary;
}

int* yytext2integers(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;

  /* Fill up ary */
  int *ary    = (int *) calloc (nary, sizeof(int));

  p = strtok(yytext,token);
  i = 0;
  while ( (p = strtok(NULL,token)) != NULL) {
    ary[i++] = atoi(p);
  }
  return ary;
}

int yytext2nintegers(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext,token);
  while ( (p = strtok(NULL,token)) != NULL) {
    nary++;
  }
  return nary;
}

int yytext2int(char* yytext, int which_token, int max_token) {
  int inumber;
  char token[] = " ";
  char *p, *s=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments on line %d\n", line_number+1);
  }
  inumber = atoi(s);
  /*  printf("inumner %d\n",inumber);*/
  return inumber;
}

char* yytext2string(char* yytext, int which_token, int max_token) {
  char token[] = " ";
  char *p, *s=NULL, *string=NULL;
  int count;

  p = strtok(yytext,token);
  count = 1;
  while ( (p = strtok(NULL,token)) != NULL) {
    count++;
    if ( count == which_token ) {
      s = p;
    }
  }
  if ( (count != which_token) && (count>max_token) ) {
    fprintf(stderr,"Wrong number of arguments on line %d\n", line_number+1);
  }
  string = (char *) calloc (strlen(s)+1, sizeof (char));
  strcpy (string, s);
  return string;
}

